_ = require 'underscore'
{events} = modules

# reserved at template
name_reserved = [
	'type',
	'bind',
	'template'
	'children'
	'repeat'
	'show'
	'max_size'
	'min_size'
	'id'
]

to_arr = (arg) ->
	if _.isString arg
		arg.split(' ').map (c) -> parseFloat c				
	else
		arg

interpolate = (str, scope) ->
	if _.isString str
		resArr = []
		recent = 0
		str.replace /{{([^}]+)}}/g, (match, p1, offset, string) ->
			if recent < offset 
				resArr.push str.substr recent, offset - recent
			recent = offset + match.length
			fn = new Function([], 'return (' + p1 + ');')
			resArr.push fn.call(scope)
			''
		if recent < str.length
			resArr.push str.substr recent, str.length - recent

		something = null
		for mola in resArr
			unless something?
				something = mola
			else
				something += mola

		something
	else
		str

genScope = (parentScope) ->
	Scope = ->
	Scope.prototype = parentScope

	$scope = new Scope()
	$scope.$$childScopes = []
	$scope.$$watches = []
	$scope.$event = new events.EventEmitter()

	$scope.$watch = (target, listener) ->
		$scope.$$watches.push
			target: (new Function([], 'return String(' + target + ');')).bind($scope)
			listener: listener
			prev: null
	 
	innerApply_r = ($scope) ->
		changed = false
	
		$scope.$$watches.forEach (w) ->
			cur = w.target()
			if w.prev != cur
				w.listener.call($scope, cur, w.prev)
				w.prev = cur
				changed = true

		_.reduce $scope.$$childScopes, ((memo, s) -> innerApply_r(s) or memo), changed
	
	$scope.$apply = (fn) ->
		fn?()

		loop
			break unless innerApply_r $scope.$rootScope
		null

	$scope.$update = ->
		root_view = $scope.$rootScope.$$view
		if root_view?
			if root_view.$update?
				root_view.$update()
			else
				root_view.$$should_update = true
		null

	$scope

module.exports = (dict) ->
	create_view_r = (tem, parentScope, ctrlers, oldTrace) ->
		trace = []

		is_root_scope = true unless parentScope?		
		is_first_creation = true unless oldTrace?
		is_scope_view = true if tem.controller?

		d = dict[tem.type]
		unless d?
			throw 'unsupported type' + tem.type

		node = null
		initArgs = []
		initArgNames = []
		if not is_first_creation
			node = oldTrace.shift()
			scope = oldTrace.shift()
		else
			# first creation
			if tem.controller?
				scope = genScope parentScope
				if is_root_scope
					scope.$rootScope = scope
					scope.$$elements = {}
				else
					parentScope.$$childScopes.push scope
				ctrl = ctrlers[tem.controller]
				# call user's controller function!!
				ctrl scope, scope.$event
			else
				scope = parentScope

			if d.cons_args?
				for arg in d.cons_args
					# arg is string... that is name of argument
					initArgs.push interpolate(tem[arg], scope)
					initArgNames.push arg
				node = d.cons initArgs...
			else
				node = d.cons()

			if is_root_scope
				throw 'the root scope must has a controller!' unless scope?
			
			if is_scope_view
				scope.$$view = node

		trace.push node
		trace.push scope
		
		# functions that would be called
		fnCalled = []
		for name of tem
			prop = tem[name]
			if not _.contains name_reserved, name
				if not _.contains initArgNames, name
					fnCalled.push name

		while d?
			for name, idx in fnCalled
				if typeof d[name] == 'function'
					# do property function...!
					d[name].call node, interpolate(tem[name], scope)
					delete fnCalled[idx]

			fnCalled = _.compact fnCalled
			d = dict[d.inherit]

		if tem.min_size?
			res = to_arr interpolate tem.min_size, scope
			bbox = node.getBoundingBox()
			x_ratio = res[0] / bbox.width
			y_ratio = res[1] / bbox.height
			max_ratio = Math.max x_ratio, y_ratio
			if max_ratio > 1
				node.setScaleX max_ratio * node.getScaleX()
				node.setScaleY max_ratio * node.getScaleY()

		if tem.max_size?
			res = to_arr interpolate tem.max_size, scope
			bbox = node.getBoundingBox()
			x_ratio = res[0] / bbox.width
			y_ratio = res[1] / bbox.height
			min_ratio = Math.min x_ratio, y_ratio
			if min_ratio < 1
				node.setScaleX min_ratio * node.getScaleX()
				node.setScaleY min_ratio * node.getScaleY()

		if tem.id?
			res = interpolate tem.id, scope
			scope.$rootScope.$$elements[res] = node

#		if tem.post_controller?
#			ctrler = ctrlers[tem.post_controller]
#			ctrler?.call(node, scope)

		if tem.show?
			res = interpolate tem.show, scope
			if res == true or res == 'true'
				node.setVisible true
			else
				node.setVisible false

		if tem.children?
			oldChildTrace = oldTrace?.shift()
			trace.push tem.children.map (c,index) ->
				[childNode, childTrace] = create_view_r(c, scope, ctrlers, oldChildTrace?[index])
				node.addChild childNode, (childNode._pri or 0) unless oldTrace?
				childTrace
			
		if tem.repeat?
			[itr,_in_,container] = tem.repeat.bind.split(' ')
			if _in_ != 'in'
				throw 'repeat usage wrong... [IDENT \'in\' IDENT]'
			src = scope[container]
			
			oldChildTrace = oldTrace?.shift()
			visited = []
			trace.push src?.map (s, index) ->
				child_scope = genScope scope
				child_scope[itr] = s
				s.$index = index
				id = index
				visited.push id

				childTraceToUse = null
				traces = oldChildTrace?[index]
				if traces?
					[old_id, childTraceToUse] = traces		
					if old_id != id
						log 'should handle this situation!'

				[childNode, childTrace] = create_view_r(tem.repeat.template, child_scope, ctrlers, childTraceToUse)
				delete s["$index"]
				node.addChild childNode, (childNode._pri or 0) unless childTraceToUse?
				[id, childTrace]

			if oldChildTrace?
				oldChildTrace.forEach (t) -> 
					unless _.contains(visited,t[0])			
						t[1][0].removeFromParent()

#		node.$update = ->
#			log 'update called'
#			[_unused_, trace] = create_view_r tem, parentScope, ctrlers, oldTrace
		
		if is_root_scope
			node.$update = ->
				create_view_r tem, null, ctrlers, trace
			if node.$$should_update == true
				node.$$should_update = false
				node.$update()

		if is_first_creation? and scope.$$view == node
			scope.$event.emit 'created', scope.$rootScope.$$elements

		[node, trace]

	create_view = (template, ctrlers) ->
		[view, trace] = create_view_r template, null, ctrlers

#		view.$update = ->
#			[_unused_, trace] = create_view_r template, null, ctrlers, trace

		view